import torch
import torch.nn as nn
import numpy as np
from deepchem.models.torch_models import layers
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.losses import L2Loss
from typing import Any


class MAT(nn.Module):
  '''An internal TorchModel class.

  In this class, we define the various layers and establish a sequential model for the Molecular Attention Transformer.
  We also define the forward call of this model in the forward function.

  References
  ----------
  .. [1] Lukasz Maziarka et al. "Molecule Attention Transformer" Graph Representation Learning workshop and Machine Learning and the Physical Sciences workshop at NeurIPS 2019. 2020. https://arxiv.org/abs/2002.08264

  Examples
  --------
  >>> import deepchem as dc
  >>> import pandas as pd
  >>> import numpy as np
  >>> smiles = ['CC', 'CCC',  'CCCC', 'CCCCC', 'CCCCCCC']
  >>> vals = [1.35, 6.72, 5.67, 1.23, 1.76]
  >>> df = pd.DataFrame(list(zip(smiles, vals)), columns = ['smiles', 'y'])
  >>> loader = dc.data.CSVLoader(tasks=['y'], feature_field='smiles', featurizer=dc.feat.MATFeaturizer())
  >>> df.to_csv('test.csv')
  >>> dataset = loader.create_dataset('test.csv')
  >>> model = dc.models.torch_models.MAT()
  >>> # To simulate input data, we will generate matrices for a single molecule.
  >>> vals = dataset.X[0]
  >>> node = vals.node_features
  >>> adj = vals.adjacency_matrix
  >>> dist = vals.distance_matrix
  >>> # We will now utilize a helper function defined in MATModel to get our matrices ready, and convert them into a batch consisting of a single molecule.
  >>> helper = dc.models.torch_models.MATModel()
  >>> node_features = helper.pad_sequence(torch.tensor(node).unsqueeze(0).float())
  >>> adjacency = helper.pad_sequence(torch.tensor(adj).unsqueeze(0).float())
  >>> distance = helper.pad_sequence(torch.tensor(dist).unsqueeze(0).float())
  >>> inputs = [node_features, adjacency, distance]
  >>> inputs = [x.astype(np.float32) if x.dtype == np.float64 else x for x in inputs]
  >>> # Get the forward call of the model for this batch.
  >>> output = model(inputs)
  '''

  def __init__(self,
               dist_kernel: str = 'softmax',
               n_encoders=8,
               lambda_attention: float = 0.33,
               lambda_distance: float = 0.33,
               h: int = 16,
               sa_hsize: int = 1024,
               sa_dropout_p: float = 0.0,
               output_bias: bool = True,
               d_input: int = 1024,
               d_hidden: int = 1024,
               d_output: int = 1024,
               activation: str = 'leakyrelu',
               n_layers: int = 1,
               ff_dropout_p: float = 0.0,
               encoder_hsize: int = 1024,
               encoder_dropout_p: float = 0.0,
               embed_input_hsize: int = 36,
               embed_dropout_p: float = 0.0,
               gen_aggregation_type: str = 'mean',
               gen_dropout_p: float = 0.0,
               gen_n_layers: int = 1,
               gen_attn_hidden: int = 128,
               gen_attn_out: int = 4,
               gen_d_output: int = 1,
               **kwargs):
    '''
    Initialization for the internal MAT class.

    Parameters
    ----------
    dist_kernel: str
        Kernel activation to be used. Can be either 'softmax' for softmax or 'exp' for exponential, for the self-attention layer.
    n_encoders: int
        Number of encoder layers in the encoder block.
    lambda_attention: float
        Constant to be multiplied with the attention matrix in the self-attention layer.
    lambda_distance: float
        Constant to be multiplied with the distance matrix in the self-attention layer.
    h: int
        Number of attention heads for the self-attention layer.
    sa_hsize: int
        Size of dense layer in the self-attention layer.
    sa_dropout_p: float
        Dropout probability for the self-attention layer.
    output_bias: bool
        If True, dense layers will use bias vectors in the self-attention layer.
    d_input: int
        Size of input layer in the feed-forward layer.
    d_hidden: int
        Size of hidden layer in the feed-forward layer. Will also be used as d_output for the MATEmbedding layer.
    d_output: int
        Size of output layer in the feed-forward layer.
    activation: str
        Activation function to be used in the feed-forward layer.
        Can choose between 'relu' for ReLU, 'leakyrelu' for LeakyReLU, 'prelu' for PReLU,
        'tanh' for TanH, 'selu' for SELU, 'elu' for ELU and 'linear' for linear activation.
    n_layers: int
        Number of layers in the feed-forward layer.
    ff_dropout_p: float
        Dropout probability in the feeed-forward layer.
    encoder_hsize: int
        Size of Dense layer for the encoder itself.
    encoder_dropout_p: float
        Dropout probability for connections in the encoder layer.
    embed_input_hsize: int
        Size of input layer for the MATEmbedding layer.
    embed_dropout_p: float
        Dropout probability for the MATEmbedding layer.
    gen_aggregation_type: str
        Type of aggregation to be used. Can be 'grover', 'mean' or 'contextual'.
    gen_dropout_p: float
        Dropout probability for the MATGenerator layer.
    gen_n_layers: int
        Number of layers in MATGenerator.
    gen_attn_hidden: int
        Size of hidden attention layer in the MATGenerator layer.
    gen_attn_out: int
        Size of output attention layer in the MATGenerator layer.
    gen_d_output: int
        Size of output layer in the MATGenerator layer.
    '''

    super(MAT, self).__init__()

    self.embedding = layers.MATEmbedding(
        d_input=embed_input_hsize, d_output=d_hidden, dropout_p=embed_dropout_p)

    self.encoder = nn.ModuleList([
        layers.MATEncoderLayer(
            dist_kernel=dist_kernel,
            lambda_attention=lambda_attention,
            lambda_distance=lambda_distance,
            h=h,
            sa_hsize=sa_hsize,
            sa_dropout_p=sa_dropout_p,
            output_bias=output_bias,
            d_input=d_input,
            d_hidden=d_hidden,
            d_output=d_output,
            activation=activation,
            n_layers=n_layers,
            ff_dropout_p=ff_dropout_p,
            encoder_hsize=encoder_hsize,
            encoder_dropout_p=encoder_dropout_p) for _ in range(n_encoders)
    ])

    self.generator = layers.MATGenerator(
        hsize=d_input,
        aggregation_type=gen_aggregation_type,
        d_output=gen_d_output,
        n_layers=gen_n_layers,
        dropout_p=gen_dropout_p,
        attn_hidden=gen_attn_hidden,
        attn_out=gen_attn_out)

  def forward(self, data: np.ndarray, **kwargs):
    node_features = torch.tensor(data[0]).float()
    adjacency_matrix = torch.tensor(data[1]).float()
    distance_matrix = torch.tensor(data[2]).float()

    mask = torch.sum(torch.abs(node_features), dim=-1) != 0
    output = self.embedding(node_features)

    for layer in self.encoder:
      output = layer(output, mask, adjacency_matrix, distance_matrix)
    output = self.generator(output, mask)
    return output


class MATModel(TorchModel):
  """Molecular Attention Transformer.

  This class implements the Molecular Attention Transformer [1]_.
  The MATFeaturizer (deepchem.feat.MATFeaturizer) is intended to work with this class.
  The model takes a batch of MATEncodings (from MATFeaturizer) as input, and returns an array of size Nx1, where N is the number of molecules in the batch.
  Each molecule is broken down into its Node Features matrix, adjacency matrix and distance matrix.
  A mask tensor is calculated for the batch. All of this goes as input to the MATEmbedding, MATEncoder and MATGenerator layers, which are defined in deepchem.models.torch_models.layers.py

  Currently, MATModel is intended to be a regression model for the freesolv dataset.

  References
  ----------
  .. [1] Lukasz Maziarka et al. "Molecule Attention Transformer" Graph Representation Learning workshop and Machine Learning and the Physical Sciences workshop at NeurIPS 2019. 2020. https://arxiv.org/abs/2002.08264

  Examples
  --------
  >>> import deepchem as dc
  >>> import pandas as pd
  >>> smiles = ['CC', 'CCC',  'CCCC', 'CCCCC', 'CCCCCCC']
  >>> vals = [1.35, 6.72, 5.67, 1.23, 1.76]
  >>> df = pd.DataFrame(list(zip(smiles, vals)), columns = ['smiles', 'y'])
  >>> loader = dc.data.CSVLoader(tasks=['y'], feature_field='smiles', featurizer=dc.feat.MATFeaturizer())
  >>> df.to_csv('test.csv')
  >>> dataset = loader.create_dataset('test.csv')
  >>> model = dc.models.torch_models.MATModel(batch_size = 2)
  >>> out = model.fit(dataset, nb_epoch = 1)
  """

  def __init__(self,
               dist_kernel: str = 'softmax',
               n_encoders=8,
               lambda_attention: float = 0.33,
               lambda_distance: float = 0.33,
               h: int = 16,
               sa_hsize: int = 1024,
               sa_dropout_p: float = 0.0,
               output_bias: bool = True,
               d_input: int = 1024,
               d_hidden: int = 1024,
               d_output: int = 1024,
               activation: str = 'leakyrelu',
               n_layers: int = 1,
               ff_dropout_p: float = 0.0,
               encoder_hsize: int = 1024,
               encoder_dropout_p: float = 0.0,
               embed_input_hsize: int = 36,
               embed_dropout_p: float = 0.0,
               gen_aggregation_type: str = 'mean',
               gen_dropout_p: float = 0.0,
               gen_n_layers: int = 1,
               gen_attn_hidden: int = 128,
               gen_attn_out: int = 4,
               gen_d_output: int = 1,
               **kwargs):
    """The wrapper class for the Molecular Attention Transformer.

    Since we are using a custom data class as input (MATEncoding), we have overriden the default_generator function from DiskDataset and customized it to work with a batch of MATEncoding classes.

    Parameters
    ----------
    dist_kernel: str
        Kernel activation to be used. Can be either 'softmax' for softmax or 'exp' for exponential, for the self-attention layer.
    n_encoders: int
        Number of encoder layers in the encoder block.
    lambda_attention: float
        Constant to be multiplied with the attention matrix in the self-attention layer.
    lambda_distance: float
        Constant to be multiplied with the distance matrix in the self-attention layer.
    h: int
        Number of attention heads for the self-attention layer.
    sa_hsize: int
        Size of dense layer in the self-attention layer.
    sa_dropout_p: float
        Dropout probability for the self-attention layer.
    output_bias: bool
        If True, dense layers will use bias vectors in the self-attention layer.
    d_input: int
        Size of input layer in the feed-forward layer.
    d_hidden: int
        Size of hidden layer in the feed-forward layer. Will also be used as d_output for the MATEmbedding layer.
    d_output: int
        Size of output layer in the feed-forward layer.
    activation: str
        Activation function to be used in the feed-forward layer.
        Can choose between 'relu' for ReLU, 'leakyrelu' for LeakyReLU, 'prelu' for PReLU,
        'tanh' for TanH, 'selu' for SELU, 'elu' for ELU and 'linear' for linear activation.
    n_layers: int
        Number of layers in the feed-forward layer.
    ff_dropout_p: float
        Dropout probability in the feeed-forward layer.
    encoder_hsize: int
        Size of Dense layer for the encoder itself.
    encoder_dropout_p: float
        Dropout probability for connections in the encoder layer.
    embed_input_hsize: int
        Size of input layer for the MATEmbedding layer.
    embed_dropout_p: float
        Dropout probability for the MATEmbedding layer.
    gen_aggregation_type: str
        Type of aggregation to be used. Can be 'grover', 'mean' or 'contextual'.
    gen_dropout_p: float
        Dropout probability for the MATGenerator layer.
    gen_n_layers: int
        Number of layers in MATGenerator.
    gen_attn_hidden: int
        Size of hidden attention layer in the MATGenerator layer.
    gen_attn_out: int
        Size of output attention layer in the MATGenerator layer.
    gen_d_output: int
        Size of output layer in the MATGenerator layer.
    """
    model = MAT(
        dist_kernel=dist_kernel,
        n_encoders=n_encoders,
        lambda_attention=lambda_attention,
        lambda_distance=lambda_distance,
        h=h,
        sa_hsize=sa_hsize,
        sa_dropout_p=sa_dropout_p,
        output_bias=output_bias,
        d_input=d_input,
        d_hidden=d_hidden,
        d_output=d_output,
        activation=activation,
        n_layers=n_layers,
        ff_dropout_p=ff_dropout_p,
        encoder_hsize=encoder_hsize,
        encoder_dropout_p=encoder_dropout_p,
        embed_input_hsize=embed_input_hsize,
        embed_dropout_p=embed_dropout_p,
        gen_aggregation_type=gen_aggregation_type,
        gen_dropout_p=gen_dropout_p,
        gen_n_layers=gen_n_layers,
        gen_attn_hidden=gen_attn_hidden,
        gen_attn_out=gen_attn_out,
        gen_d_output=gen_d_output)

    loss = L2Loss()
    output_types = ['prediction']
    super(MATModel, self).__init__(
        model, loss=loss, output_types=output_types, **kwargs)

  def pad_array(self, array: np.ndarray, shape: Any) -> np.ndarray:
    """
    Pads an array to the desired shape.

    Parameters
    ----------
    array: np.ndarray
    Array to be padded.
    shape: int or Tuple
    Shape the array is padded to.

    Returns
    ----------
    array: np.ndarray
    Array padded to input shape.
    """
    result = np.zeros(shape=shape)
    slices = tuple(slice(s) for s in array.shape)
    result[slices] = array
    return result

  def pad_sequence(self, sequence: np.ndarray) -> np.ndarray:
    """
    Pads a given sequence using the pad_array function.

    Parameters
    ----------
    sequence: np.ndarray
    Arrays in this sequence are padded to the largest shape in the sequence.

    Returns
    ----------
    array: np.ndarray
    Sequence with padded arrays.
    """
    shapes = np.stack([np.array(t.shape) for t in sequence])
    max_shape = tuple(np.max(shapes, axis=0))
    return np.stack([self.pad_array(t, shape=max_shape) for t in sequence])

  def default_generator(self,
                        dataset,
                        epochs=1,
                        mode='fit',
                        deterministic=True,
                        pad_batches=True,
                        **kwargs):
    for epoch in range(epochs):
      for (X_b, y_b, w_b, ids_b) in dataset.iterbatches(
          batch_size=self.batch_size,
          deterministic=deterministic,
          pad_batches=pad_batches):

        node_features = self.pad_sequence(
            [torch.tensor(data.node_features).float() for data in X_b])
        adjacency_matrix = self.pad_sequence(
            [torch.tensor(data.adjacency_matrix).float() for data in X_b])
        distance_matrix = self.pad_sequence(
            [torch.tensor(data.distance_matrix).float() for data in X_b])

        inputs = [node_features, adjacency_matrix, distance_matrix]
        yield (inputs, [y_b], [w_b])



"""
Test for fake_data_generator.py
"""
import numpy as np
from deepchem.utils.fake_data_generator import FakeGraphGenerator, generate_edge_index, remove_self_loops


def test_fake_graph_dataset():
  n_graphs = 10
  n_node_features = 5
  n_edge_features = 3
  n_classes = 2
  z_shape = 5

  # graph-level labels
  fgg = FakeGraphGenerator(
      min_nodes=3,
      max_nodes=10,
      n_node_features=n_node_features,
      avg_degree=4,
      n_edge_features=n_edge_features,
      n_classes=n_classes,
      task='graph',
      z=z_shape)
  graphs = fgg.sample(n_graphs=n_graphs)

  assert len(graphs) == n_graphs
  assert np.unique(graphs.y).shape == (n_classes,)

  graph = graphs.X[0]
  assert graph.node_features.shape[1] == n_node_features
  assert graph.edge_features.shape[1] == n_edge_features
  assert graph.z.shape == (1, z_shape)

  # node-level labels
  fgg = FakeGraphGenerator(
      min_nodes=3,
      max_nodes=10,
      n_node_features=n_node_features,
      avg_degree=4,
      n_edge_features=n_edge_features,
      n_classes=n_classes,
      task='node',
      z=z_shape)
  graphs = fgg.sample(n_graphs=n_graphs)

  assert len(graphs) == n_graphs

  graph = graphs.X[0]
  # graph.y contains node-labels and graph.node_features.shape[0]
  # holds number of nodes in that graph
  assert graph.y.shape[0] == graph.node_features.shape[0]
  assert graph.node_features.shape[1] == n_node_features
  assert graph.edge_features.shape[1] == n_edge_features
  assert graph.z.shape == (1, z_shape)


def test_generate_edge_index():
  n_nodes, avg_degree = 5, 3
  edge_indices = generate_edge_index(n_nodes, avg_degree, remove_loops=False)
  assert edge_indices.shape[0] == 2
  assert edge_indices.shape[1] == n_nodes * avg_degree


def test_remove_self_loops():
  edge_indices = np.array([[1, 2, 3], [1, 2, 4]])
  edge_indices = remove_self_loops(edge_indices)
  assert edge_indices.shape[0] == 2
  assert edge_indices.shape[1] == 1

  edge_indices = np.ones((2, 3))
  edge_indices = remove_self_loops(edge_indices)
  assert edge_indices.shape[0] == 2
  assert edge_indices.shape[1] == 0



"""
Topological fingerprints for macromolecular structures.
"""
import numpy as np
import logging
import itertools
from deepchem.utils.hash_utils import hash_ecfp
from deepchem.feat import ComplexFeaturizer
from deepchem.utils.rdkit_utils import load_complex
from deepchem.utils.hash_utils import vectorize
from deepchem.utils.voxel_utils import voxelize
from deepchem.utils.voxel_utils import convert_atom_to_voxel
from deepchem.utils.rdkit_utils import compute_all_ecfp
from deepchem.utils.rdkit_utils import compute_contact_centroid
from deepchem.utils.rdkit_utils import MoleculeLoadException
from deepchem.utils.geometry_utils import compute_pairwise_distances
from deepchem.utils.geometry_utils import subtract_centroid

from typing import Tuple, Dict, List

logger = logging.getLogger(__name__)


def featurize_contacts_ecfp(
    frag1: Tuple,
    frag2: Tuple,
    pairwise_distances: np.ndarray = None,
    cutoff: float = 4.5,
    ecfp_degree: int = 2) -> Tuple[Dict[int, str], Dict[int, str]]:
  """Computes ECFP dicts for pairwise interaction between two molecular fragments.

  Parameters
  ----------
  frag1: Tuple
    A tuple of (coords, mol) returned by `load_molecule`.
  frag2: Tuple
    A tuple of (coords, mol) returned by `load_molecule`.
  pairwise_distances: np.ndarray
    Array of pairwise fragment-fragment distances (Angstroms)
  cutoff: float
    Cutoff distance for contact consideration
  ecfp_degree: int
    ECFP radius

  Returns
  -------
  Tuple of dictionaries of ECFP contact fragments
  """
  if pairwise_distances is None:
    pairwise_distances = compute_pairwise_distances(frag1[0], frag2[0])
  # contacts is of form (x_coords, y_coords), a tuple of 2 lists
  contacts = np.nonzero((pairwise_distances < cutoff))
  # contacts[0] is the x_coords, that is the frag1 atoms that have
  # nonzero contact.
  frag1_atoms = set([int(c) for c in contacts[0].tolist()])
  # contacts[1] is the y_coords, the frag2 atoms with nonzero contacts
  frag2_atoms = set([int(c) for c in contacts[1].tolist()])

  frag1_ecfp_dict = compute_all_ecfp(
      frag1[1], indices=frag1_atoms, degree=ecfp_degree)
  frag2_ecfp_dict = compute_all_ecfp(
      frag2[1], indices=frag2_atoms, degree=ecfp_degree)

  return (frag1_ecfp_dict, frag2_ecfp_dict)


class ContactCircularFingerprint(ComplexFeaturizer):
  """Compute (Morgan) fingerprints near contact points of macromolecular complexes.

  Given a macromolecular complex made up of multiple
  constituent molecules, first compute the contact points where
  atoms from different molecules come close to one another. For
  atoms within "contact regions," compute radial "ECFP"
  fragments which are sub-molecules centered at atoms in the
  contact region.

  For a macromolecular complex, returns a vector of shape
  `(2*size,)`
  """

  def __init__(self, cutoff: float = 4.5, radius: int = 2, size: int = 8):
    """
    Parameters
    ----------
    cutoff: float (default 4.5)
      Distance cutoff in angstroms for molecules in complex.
    radius: int, optional (default 2)
      Fingerprint radius.
    size: int, optional (default 8)
      Length of generated bit vector.
    """
    self.cutoff = cutoff
    self.radius = radius
    self.size = size

  def _featurize(self, datapoint, **kwargs):
    """
    Compute featurization for a molecular complex

    Parameters
    ----------
    datapoint: Tuple[str, str]
      Filenames for molecule and protein.
    """
    if 'complex' in kwargs:
      datapoint = kwargs.get("complex")
      raise DeprecationWarning(
          'Complex is being phased out as a parameter, please pass "datapoint" instead.'
      )

    try:
      fragments = load_complex(datapoint, add_hydrogens=False)

    except MoleculeLoadException:
      logger.warning("This molecule cannot be loaded by Rdkit. Returning None")
      return None
    pairwise_features = []
    # We compute pairwise contact fingerprints
    for (frag1, frag2) in itertools.combinations(fragments, 2):
      # Get coordinates
      distances = compute_pairwise_distances(frag1[0], frag2[0])
      vector = [
          vectorize(hash_ecfp, feature_dict=ecfp_dict, size=self.size)
          for ecfp_dict in featurize_contacts_ecfp(
              frag1,
              frag2,
              distances,
              cutoff=self.cutoff,
              ecfp_degree=self.radius)
      ]
      pairwise_features += vector

    pairwise_features = np.concatenate(pairwise_features)
    return pairwise_features


class ContactCircularVoxelizer(ComplexFeaturizer):
  """Computes ECFP fingerprints on a voxel grid.

  Given a macromolecular complex made up of multiple
  constituent molecules, first compute the contact points where
  atoms from different molecules come close to one another. For
  atoms within "contact regions," compute radial "ECFP"
  fragments which are sub-molecules centered at atoms in the
  contact region. Localize these ECFP fingeprints at the voxel
  in which they originated.

  Featurizes a macromolecular complex into a tensor of shape
  `(voxels_per_edge, voxels_per_edge, voxels_per_edge, size)` where
  `voxels_per_edge = int(box_width/voxel_width)`. If `flatten==True`,
  then returns a flattened version of this tensor of length
  `size*voxels_per_edge**3`
  """

  def __init__(self,
               cutoff: float = 4.5,
               radius: int = 2,
               size: int = 8,
               box_width: float = 16.0,
               voxel_width: float = 1.0,
               flatten: bool = False):
    """
    Parameters
    ----------
    cutoff: float (default 4.5)
      Distance cutoff in angstroms for molecules in complex.
    radius : int, optional (default 2)
      Fingerprint radius.
    size : int, optional (default 8)
      Length of generated bit vector.
    box_width: float, optional (default 16.0)
      Size of a box in which voxel features are calculated. Box
      is centered on a ligand centroid.
    voxel_width: float, optional (default 1.0)
      Size of a 3D voxel in a grid.
    flatten: bool, optional (default False)
      If True, then returns a flat feature vector rather than voxel grid. This
      feature vector is constructed by flattening the usual voxel grid.
    """
    self.cutoff = cutoff
    self.radius = radius
    self.size = size
    self.box_width = box_width
    self.voxel_width = voxel_width
    self.voxels_per_edge = int(self.box_width / self.voxel_width)
    self.flatten = flatten

  def _featurize(self, complex):
    """
    Compute featurization for a molecular complex

    Parameters
    ----------
    complex: Tuple[str, str]
      Filenames for molecule and protein.
    """
    try:
      fragments = load_complex(complex, add_hydrogens=False)

    except MoleculeLoadException:
      logger.warning("This molecule cannot be loaded by Rdkit. Returning None")
      return None
    pairwise_features: List[np.ndarray] = []
    # We compute pairwise contact fingerprints
    centroid = compute_contact_centroid(fragments, cutoff=self.cutoff)
    for (frag1, frag2) in itertools.combinations(fragments, 2):
      distances = compute_pairwise_distances(frag1[0], frag2[0])
      frag1_xyz = subtract_centroid(frag1[0], centroid)
      frag2_xyz = subtract_centroid(frag2[0], centroid)
      xyzs = [frag1_xyz, frag2_xyz]
      pairwise_features.append(
          sum([
              voxelize(
                  convert_atom_to_voxel,
                  xyz,
                  self.box_width,
                  self.voxel_width,
                  hash_function=hash_ecfp,
                  feature_dict=ecfp_dict,
                  nb_channel=self.size) for xyz, ecfp_dict in zip(
                      xyzs,
                      featurize_contacts_ecfp(
                          frag1,
                          frag2,
                          distances,
                          cutoff=self.cutoff,
                          ecfp_degree=self.radius))
          ]))
    if self.flatten:
      return np.concatenate(
          [features.flatten() for features in pairwise_features])
    else:
      # Features are of shape (voxels_per_edge, voxels_per_edge,
      # voxels_per_edge, num_feat) so we should concatenate on the last
      # axis.
      return np.concatenate(pairwise_features, axis=-1)


def compute_all_sybyl(mol, indices=None):
  """Computes Sybyl atom types for atoms in molecule."""
  raise NotImplementedError("This function is not implemented yet")


def featurize_binding_pocket_sybyl(protein_xyz,
                                   protein,
                                   ligand_xyz,
                                   ligand,
                                   pairwise_distances=None,
                                   cutoff=7.0):
  """Computes Sybyl dicts for ligand and binding pocket of the protein.

  Parameters
  ----------
  protein_xyz: np.ndarray
    Of shape (N_protein_atoms, 3)
  protein: Rdkit Molecule
    Contains more metadata.
  ligand_xyz: np.ndarray
    Of shape (N_ligand_atoms, 3)
  ligand: Rdkit Molecule
    Contains more metadata
  pairwise_distances: np.ndarray
    Array of pairwise protein-ligand distances (Angstroms)
  cutoff: float
    Cutoff distance for contact consideration.
  """

  if pairwise_distances is None:
    pairwise_distances = compute_pairwise_distances(protein_xyz, ligand_xyz)
  contacts = np.nonzero((pairwise_distances < cutoff))
  protein_atoms = set([int(c) for c in contacts[0].tolist()])

  protein_sybyl_dict = compute_all_sybyl(protein, indices=protein_atoms)
  ligand_sybyl_dict = compute_all_sybyl(ligand)
  return (protein_sybyl_dict, ligand_sybyl_dict)



import numpy as np
from deepchem.feat import GraphData
from typing import List, Sequence, Optional


class _MapperDMPNN:
  """
  This class is a helper class for DMPNNModel class to generate concatenated feature vector and mappings.

  `self.f_ini_atoms_bonds` is the concatenated feature vector which contains
  concatenation of initial atom and bond features.

  `self.atom_to_incoming_bonds` is mapping from atom index to list of indicies of incoming bonds.

  `self.mapping` is the mapping that maps bond index to 'array of indices of the bonds'
  incoming at the initial atom of the bond (excluding the reverse bonds)

  Example,
  Let the diagram given below represent a molecule containing 3 atoms (nodes) and 2 bonds (edges):-

  |   0 --- 1
  |   |
  |   2

  Here, atoms are => A0, A1 and A2 and their respective feature vectors are f0, f1, and f2.

  Let the bonds from atoms 0->1 ('B[01]') and 1->0 ('B[10]') be considered as 2 different bonds.
  Hence, by considering the same for all atoms, the total number of bonds = 4.

  Let:
    B[01] => B0
    B[10] => B1
    B[02] => B2
    B[20] => B3

  Hence repective bond features are fb0, fb1, fb2, and fb3.
  (Note: fb0 = fb1, fb2 = fb3)

  'f_ini_atoms_bonds' is the concatenated feature vector which contains
  concatenation of initial atom and bond features.

                   'B0'
    Example: 'A0' -----> A1 , concat feature = f0 + fb0

  Hence,
                            B0       B1       B2       B3     B(-1)
    f_ini_atoms_bonds = [ f0+fb0 , f1+fb1 , f0+fb2 , f2+fb3 , f(-1) ]

    (Note: f(-1) is a zero array of the same size as other concatenated features.)

  `atom_to_incoming_bonds` is mapping from atom index to list of indicies of incoming bonds.

                   B3          B1
    Example: 'A2' ----> 'A0' <---- 'A1', for A0 => [B1, B3]

  Hence,
                                 A0        A1        A2
    atom_to_incoming_bonds = [ [B1,B3] [B0,B(-1)] [B2,B(-1)] ]

    (Note: Here, maximum number of incoming bonds is 2. So, -1 index is added to all those cases
           where number of incoming bonds is less than maximum. In this case, its for A1 and A2.)

  To get mapping, first find indices of the bonds, incoming at the initial atom of the bond.

    Example: for bond B0, B1 and B3 are coming towards atom 0.

  |                    B0             B1
  |                 0 ----> 1  |  0 <---- 1
  |                            |  ^
  |                            |  | B3
  |                            |  2

                                     B0     B1    B2     B3
  mapping (with reverse bonds) = [ [B1,B3] [B0] [B1,B3] [B2] ]

  To get the required mapping, reverse bond indices are replaced with -1
  and extra space in the array elements is filled with -1, to get a uniform array.

  The mapping is also padded with -1 at the end, so that the length of `mapping` is
  equal to the length of `f_ini_atoms_bonds`.

  Hence,
                    B0          B1          B2            B3          B(-1)
    mapping = [ [B(-1),B3] [B(-1),B(-1)] [B1,B(-1)] [B(-1),B(-1)] [B(-1),B(-1)] ]

    OR

    mapping = [[-1, 3], [-1, -1], [1, -1], [-1, -1], [-1, -1]]
  """

  def __init__(self, graph: GraphData):
    """
    Parameters
    ----------
    graph: GraphData
      GraphData object.
    """
    self.num_atoms: int = graph.num_nodes
    self.num_atom_features: int = graph.num_node_features
    self.num_bonds: int = graph.num_edges
    self.num_bond_features: int = graph.num_edge_features
    self.atom_features: np.ndarray = graph.node_features
    self.bond_features: Optional[np.ndarray] = graph.edge_features
    self.bond_index: np.ndarray = graph.edge_index
    self.global_features: np.ndarray = graph.global_features  # type: ignore
    # mypy check is ignored for global_features as it is not a default attribute
    # of GraphData. It is created during runtime using **kwargs.

    # mapping from bond index to the index of the atom (where the bond is coming from)
    self.bond_to_ini_atom: np.ndarray

    # mapping from bond index to concat(in_atom, bond) features
    self.f_ini_atoms_bonds: np.ndarray = np.empty(0)

    # mapping from atom index to list of indicies of incoming bonds
    self.atom_to_incoming_bonds: np.ndarray

    # mapping which maps bond index to 'array of indices of the bonds' incoming at the initial atom of the bond (excluding the reverse bonds)
    self.mapping: np.ndarray = np.empty(0)

    if self.num_bonds == 0:
      self.bond_to_ini_atom = np.empty(0)
      self.f_ini_atoms_bonds = np.zeros(
          (1, self.num_atom_features + self.num_bond_features))

      self.atom_to_incoming_bonds = np.asarray([[-1]], dtype=int)
      self.mapping = np.asarray([[-1]], dtype=int)

    else:
      self.bond_to_ini_atom = self.bond_index[0]
      self._get_f_ini_atoms_bonds()  # its zero padded at the end

      self.atom_to_incoming_bonds = self._get_atom_to_incoming_bonds()
      self._generate_mapping()  # its padded with -1 at the end

  @property
  def values(self) -> Sequence[np.ndarray]:
    """
    Returns the required mappings:
    - atom features
    - concat features (atom + bond)
    - atom to incoming bonds mapping
    - mapping
    - global features
    """
    return self.atom_features, self.f_ini_atoms_bonds, self.atom_to_incoming_bonds, self.mapping, self.global_features

  def _get_f_ini_atoms_bonds(self):
    """
    Method to get `self.f_ini_atoms_bonds`
    """
    self.f_ini_atoms_bonds = np.hstack(
        (self.atom_features[self.bond_to_ini_atom], self.bond_features))

    # zero padded at the end
    self.f_ini_atoms_bonds = np.pad(self.f_ini_atoms_bonds, ((0, 1), (0, 0)))

  def _generate_mapping(self):
    """
    Generate mapping, which maps bond index to 'array of indices of the bonds'
    incoming at the initial atom of the bond (reverse bonds are not considered).

    Steps:
    - Get mapping based on `self.atom_to_incoming_bonds` and `self.bond_to_ini_atom`.
    - Replace reverse bond indices with -1.
    - Pad the mapping with -1.
    """

    # get mapping which maps bond index to 'array of indices of the bonds' incoming at the initial atom of the bond
    self.mapping = self.atom_to_incoming_bonds[self.bond_to_ini_atom]
    self._replace_rev_bonds()

    # padded with -1 at the end
    self.mapping = np.pad(self.mapping, ((0, 1), (0, 0)), constant_values=-1)

  def _get_atom_to_incoming_bonds(self) -> np.ndarray:
    """
    Method to get atom_to_incoming_bonds mapping
    """
    # mapping from bond index to the index of the atom (where the bond if going to)
    bond_to_final_atom: np.ndarray = self.bond_index[1]

    # mapping from atom index to list of indicies of incoming bonds
    a2b: List = []
    for i in range(self.num_atoms):
      a2b.append(list(np.where(bond_to_final_atom == i)[0]))

    # get maximum number of incoming bonds
    max_num_bonds: int = max(1,
                             max(len(incoming_bonds) for incoming_bonds in a2b))

    # Make number of incoming bonds equal to maximum number of bonds.
    # This is done by appending -1 to fill remaining space at each atom indices.
    a2b = [
        a2b[a] + [-1] * (max_num_bonds - len(a2b[a]))
        for a in range(self.num_atoms)
    ]

    return np.asarray(a2b, dtype=int)

  def _replace_rev_bonds(self):
    """
    Method to get b2revb and replace the reverse bond indices with -1 in mapping.
    """
    # mapping from bond index to the index of the reverse bond
    b2revb: np.ndarray = np.empty(self.num_bonds, dtype=int)
    for i in range(self.num_bonds):
      if i % 2 == 0:
        b2revb[i] = i + 1
      else:
        b2revb[i] = i - 1

    for count, i in enumerate(b2revb):
      self.mapping[count][np.where(self.mapping[count] == i)] = -1



from deepchem.feat import Featurizer
from typing import Dict, List
try:
  from transformers import RobertaTokenizerFast
except ModuleNotFoundError:
  raise ImportError(
      'Transformers must be installed for RobertaFeaturizer to be used!')
  pass


class RobertaFeaturizer(RobertaTokenizerFast, Featurizer):
  """Roberta Featurizer.

  The Roberta Featurizer is a wrapper class of the Roberta Tokenizer,
  which is used by Huggingface's transformers library for tokenizing large corpuses for Roberta Models.
  Please confirm the details in [1]_.

  Please see https://github.com/huggingface/transformers
  and https://github.com/seyonechithrananda/bert-loves-chemistry for more details.

  Examples
  --------
  >>> from deepchem.feat import RobertaFeaturizer
  >>> smiles = ["Cn1c(=O)c2c(ncn2C)n(C)c1=O", "CC(=O)N1CN(C(C)=O)C(O)C1O"]
  >>> featurizer = RobertaFeaturizer.from_pretrained("seyonec/SMILES_tokenized_PubChem_shard00_160k")
  >>> out = featurizer.featurize(smiles, add_special_tokens=True, truncation=True)

  References
  ----------
  .. [1] Chithrananda, Seyone, Grand, Gabriel, and Ramsundar, Bharath (2020): "Chemberta: Large-scale self-supervised
    pretraining for molecular property prediction." arXiv. preprint. arXiv:2010.09885.


  Note
  -----
  This class requires transformers to be installed.
  RobertaFeaturizer uses dual inheritance with RobertaTokenizerFast in Huggingface for rapid tokenization,
  as well as DeepChem's MolecularFeaturizer class.
  """

  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return

  def _featurize(self, datapoint: str, **kwargs) -> List[List[int]]:
    """Calculate encoding using HuggingFace's RobertaTokenizerFast

    Parameters
    ----------
    sequence: str
      Arbitrary string sequence to be tokenized.

    Returns
    -------
    datapoint: List
      List containing two lists; the `input_ids` and the `attention_mask`
    """

    # the encoding is natively a dictionary with keys 'input_ids' and 'attention_mask'
    encoding = list(self(datapoint, **kwargs).values())
    return encoding

  def __call__(self, *args, **kwargs) -> Dict[str, List[int]]:
    return super().__call__(*args, **kwargs)



"""
A fake data generator
"""
import random
import numpy as np
from deepchem.data import NumpyDataset
from deepchem.feat import GraphData


class FakeGraphGenerator:
  """Generates a random graphs which can be used for testing or other purposes.

  The generated graph supports both node-level and graph-level labels.

  Example
  -------
  >>> from deepchem.utils.fake_data_generator import FakeGraphGenerator
  >>> fgg  = FakeGraphGenerator(min_nodes=8, max_nodes=10,  n_node_features=5, avg_degree=8, n_edge_features=3, n_classes=2, task='graph', z=5)
  >>> graphs = fgg.sample(n_graphs=10)
  >>> type(graphs)
  <class 'deepchem.data.datasets.NumpyDataset'>
  >>> type(graphs.X[0])
  <class 'deepchem.feat.graph_data.GraphData'>
  >>> len(graphs) == 10  # num_graphs
  True

  Note
  ----
  The FakeGraphGenerator class is based on torch_geometric.dataset.FakeDataset
  class.
  """

  def __init__(self,
               min_nodes: int = 10,
               max_nodes: int = 10,
               n_node_features: int = 5,
               avg_degree: int = 4,
               n_edge_features: int = 3,
               n_classes: int = 2,
               task: str = 'graph',
               **kwargs):
    """
    Parameters
    ----------
    min_nodes: int, default 10
      Minimum number of permissible nodes in a graph
    max_nodes: int, default 10
      Maximum number of permissible nodes in a graph
    n_node_features: int, default 5
      Average number of node features in a graph
    avg_degree: int, default 4
      Average degree of the graph (avg_degree should be a positive number greater than the min_nodes)
    n_edge_features: int, default 3
      Average number of features in the edge
    task: str, default 'graph'
      Indicates node-level labels or graph-level labels
    kwargs: optional
      Additional graph attributes and their shapes , e.g. `global_features = 5`
    """
    assert avg_degree >= 1, "Average degree should be greater than 0"
    self.min_nodes = min_nodes
    self.max_nodes = max_nodes
    self.avg_degree = avg_degree
    self.n_node_features = n_node_features
    self.n_edge_features = n_edge_features
    self.n_classes = n_classes
    self.task = task
    self.kwargs = kwargs

  def sample(self, n_graphs: int = 100) -> NumpyDataset:
    """Samples graphs

    Parameters
    ----------
    n_graphs: int, default 100
      Number of graphs to generate

    Returns
    -------
    graphs: NumpyDataset
      Generated Graphs
    """
    graphs, labels = [], []
    for i in range(n_graphs):
      n_nodes = random.randint(self.min_nodes, self.max_nodes)
      edge_index = generate_edge_index(n_nodes, self.avg_degree)
      n_edges = edge_index.shape[1]

      if self.task == 'graph':
        graph_label = random.randint(0, self.n_classes - 1)
        node_features = np.random.rand(n_nodes,
                                       self.n_node_features) + graph_label
        edge_features = np.random.rand(n_edges,
                                       self.n_edge_features) + graph_label
        kwargs = {}
        for feature_name, feature_shape in self.kwargs.items():
          kwargs[feature_name] = np.random.rand(1, feature_shape) + graph_label
        labels.append(graph_label)

      elif self.task == 'node':
        node_label = np.random.randint(0, self.n_classes - 1,
                                       n_nodes).astype(np.float64)
        node_features = np.random.rand(
            n_nodes, self.n_node_features) + node_label.reshape(-1, 1)
        # For a node-prediction task, label is not added to edge features and other global features
        # because label here is a node-level attribute and not a graph-level attribute
        edge_features = np.random.rand(n_edges, self.n_edge_features)
        kwargs = {}
        for feature_name, feature_shape in self.kwargs.items():
          kwargs[feature_name] = np.random.rand(1, feature_shape)
        kwargs['y'] = node_label

      graph = GraphData(node_features, edge_index, edge_features, **kwargs)
      graphs.append(graph)

      if self.task == 'graph':
        graph_dataset = NumpyDataset(X=np.array(graphs), y=np.array(labels))
      elif self.task == 'node':
        # In this case, the 'y' attribute of GraphData will contain the
        # node-level labels.
        graph_dataset = NumpyDataset(X=np.array(graphs))
    return graph_dataset


def generate_edge_index(n_nodes: int,
                        avg_degree: int,
                        remove_loops: bool = True) -> np.ndarray:
  """Returns source and destination nodes for `num_nodes * avg_degree` number of randomly
  generated edges. If remove_loops is True, then self-loops from the edge_index pairs
  are removed.

  Parameters
  ----------
  n_nodes: int
    Number of nodes in the graph
  avg_degree: int
    Average degree per node in a graph
  remove_loops: bool
    Remove self-loops in a graph
  """
  n_edges = n_nodes * avg_degree
  edge_index = np.random.randint(low=0, high=n_nodes, size=(2, n_edges))

  if remove_loops:
    edge_index = remove_self_loops(edge_index)
  return edge_index


def remove_self_loops(edge_index: np.ndarray) -> np.ndarray:
  """Removes self-loops from a given set of edges

  Parameters
  ----------
  edge_index: numpy.ndarray
    An numpy array of shape (2, |num_edges|) representing edges in a graph
  """
  mask = []
  for i in range(edge_index.shape[1]):
    if edge_index[0][i] != edge_index[1][i]:
      # not a self-loop
      mask.append(i)
  return edge_index[:, mask]



import numpy as np
from deepchem.feat.molecule_featurizers import OneHotFeaturizer
from deepchem.feat.base_classes import Featurizer
from typing import List, Optional

CHARSET = [
    'A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R',
    'S', 'T', 'V', 'W', 'Y', 'X', 'Z', 'B', 'U', 'O'
]


class PFMFeaturizer(Featurizer):
  """
    Encodes a list position frequency matrices for a given list of multiple sequence alignments

    The default character set is 25 amino acids. If you want to use a different character set, such as nucleotides, simply pass in
    a list of character strings in the featurizer constructor.

    The max_length parameter is the maximum length of the sequences to be featurized. If you want to featurize longer sequences, modify the
    max_length parameter in the featurizer constructor.

    The final row in the position frequency matrix is the unknown set, if there are any characters which are not included in the charset.

    Examples
    --------
    >>> from deepchem.feat.sequence_featurizers import PFMFeaturizer
    >>> from deepchem.data import NumpyDataset
    >>> msa = NumpyDataset(X=[['ABC','BCD'],['AAA','AAB']], ids=[['seq01','seq02'],['seq11','seq12']])
    >>> seqs = msa.X
    >>> featurizer = PFMFeaturizer()
    >>> pfm = featurizer.featurize(seqs)
    >>> pfm.shape
    (2, 26, 100)

    """

  def __init__(self,
               charset: List[str] = CHARSET,
               max_length: Optional[int] = 100):
    """Initialize featurizer.

        Parameters
        ----------
        charset: List[str] (default CHARSET)
            A list of strings, where each string is length 1 and unique.
        max_length: int, optional (default 25)
            Maximum length of sequences to be featurized.
        """
    if len(charset) != len(set(charset)):
      raise ValueError("All values in charset must be unique.")
    self.charset = charset
    self.max_length = max_length
    self.ohe = OneHotFeaturizer(charset=CHARSET, max_length=max_length)

  def _featurize(self, datapoint):
    """Featurize a multisequence alignment into a position frequency matrix

        Use dc.utils.sequence_utils.hhblits or dc.utils.sequence_utils.hhsearch to create a multiple sequence alignment from a fasta file.

        Parameters
        ----------
        datapoint: np.ndarray
            MSA to featurize. A list of sequences which have been aligned and padded to the same length.

        Returns
        -------
        pfm: np.ndarray
            Position frequency matrix for the set of sequences with the rows corresponding to the unique characters and the columns corresponding to the position in the alignment.

        """

    seq_one_hot = self.ohe.featurize(datapoint)

    seq_one_hot_array = np.transpose(
        np.array(seq_one_hot), (0, 2, 1)
    )  # swap rows and columns to make rows the characters, columns the positions

    pfm = np.sum(seq_one_hot_array, axis=0)

    return pfm


def PFM_to_PPM(pfm):
  """
    Calculate position probability matrix from a position frequency matrix
    """
  ppm = pfm.copy()
  for col in range(ppm.shape[1]):
    total_count = np.sum(ppm[:, col])
    if total_count > 0:
      # Calculate frequency
      ppm[:, col] = ppm[:, col] / total_count
  return ppm



from unittest import TestCase
import numpy as np

import deepchem.rl.envs.tictactoe


class TestTicTacToeEnvironment(TestCase):

  def test_constructor(self):
    board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
    assert len(board.state) == 1
    assert board.state[0].shape == (3, 3, 2)
    assert np.sum(board.state[0]) == 1 or np.sum(board.state[0]) == 0

  def test_step(self):
    board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
    X = deepchem.rl.envs.tictactoe.TicTacToeEnvironment.X
    board._state = [np.zeros(shape=(3, 3, 2), dtype=np.float32)]
    board.step(0)
    assert np.all(board.state[0][0][0] == X)

  def test_winner(self):
    board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
    X = deepchem.rl.envs.tictactoe.TicTacToeEnvironment.X
    board.state[0][0][0] = X
    board.state[0][0][1] = X
    assert not board.check_winner(X)
    board.state[0][0][2] = X
    assert board.check_winner(X)

  def test_game_over(self):
    board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
    X = deepchem.rl.envs.tictactoe.TicTacToeEnvironment.X
    board.state[0][0][0] = X
    board.state[0][0][1] = X
    assert not board.check_winner(X)
    board.state[0][0][2] = X
    assert board.check_winner(X)

  def test_display(self):
    board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
    s = board.display()
    assert s.find("X") == -1

  def test_get_O_move(self):
    board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
    empty = deepchem.rl.envs.tictactoe.TicTacToeEnvironment.EMPTY
    move = board.get_O_move()
    assert np.all(board.state[0][move[0]][move[1]] == empty)



"""Normalizing flows for transforming probability distributions using PyTorch.
"""
import torch
from torch import nn
from typing import Sequence, Tuple


class NormalizingFlow(nn.Module):
  """Normalizing flows are widley used to perform generative models.
  This algorithm gives advantages over variational autoencoders (VAE) because
  of ease in sampling by applying invertible transformations
  (Frey, Gadepally, & Ramsundar, 2022).

  Example
  --------
  >>> import deepchem as dc
  >>> from deepchem.models.torch_models.layers import Affine
  >>> from deepchem.models.torch_models.normalizing_flows_pytorch import NormalizingFlow
  >>> import torch
  >>> from torch.distributions import MultivariateNormal
  >>> # initialize the transformation layer's parameters
  >>> dim = 2
  >>> samples = 96
  >>> transforms = [Affine(dim)]
  >>> distribution = MultivariateNormal(torch.zeros(dim), torch.eye(dim))
  >>> # initialize normalizing flow model
  >>> model = NormalizingFlow(transforms, distribution, dim)
  >>> # evaluate the log_prob when applying the transformation layers
  >>> input = distribution.sample(torch.Size((samples, dim)))
  >>> len(model.log_prob(input))
  96
  >>> # evaluates the the sampling method and its log_prob
  >>> len(model.sample(samples))
  2

  """

  def __init__(self, transform: Sequence, base_distribution, dim: int) -> None:
    """This class considers a transformation, or a composition of transformations
    functions (layers), between a base distribution and a target distribution.

    Parameters
    ----------
    transform: Sequence
      Bijective transformation/transformations which are considered the layers
      of a Normalizing Flow model.
    base_distribution: torch.Tensor
      Probability distribution to initialize the algorithm. The Multivariate Normal
      distribution is mainly used for this parameter.
    dim: int
      Value of the Nth dimension of the dataset.

    """
    super().__init__()
    self.dim = dim
    self.transforms = nn.ModuleList(transform)
    self.base_distribution = base_distribution

  def log_prob(self, inputs: torch.Tensor) -> torch.Tensor:
    """This method computes the probability of the inputs when
    transformation/transformations are applied.

    Parameters
    ----------
    inputs: torch.Tensor
      Tensor used to evaluate the log_prob computation of the learned
      distribution.
      shape: (samples, dim)

    Returns
    -------
    log_prob: torch.Tensor
      This tensor contains the value of the log probability computed.
      shape: (samples)

    """
    log_prob = torch.zeros(inputs.shape[0])
    for biject in reversed(self.transforms):
      inputs, inverse_log_det_jacobian = biject.inverse(inputs)
      log_prob += inverse_log_det_jacobian

    return log_prob

  def sample(self, n_samples: int) -> Tuple[torch.Tensor, torch.Tensor]:
    """Performs a sampling from the transformed distribution.
    Besides the outputs (sampling), this method returns the logarithm of
    probability to obtain the outputs at the base distribution.

    Parameters
    ----------
    n_samples: int
      Number of samples to select from the transformed distribution

    Returns
    -------
    sample: tuple
      This tuple contains a two torch.Tensor objects. The first represents
      a sampling of the learned distribution when transformations had been
      applied. The secong torc.Tensor is the computation of log probabilities
      of the transformed distribution.
      shape: ((samples, dim), (samples))

    """
    outputs = self.base_distribution.sample((n_samples,))
    log_prob = self.base_distribution.log_prob(outputs)

    for biject in self.transforms:
      outputs, log_det_jacobian = biject.forward(outputs)
      log_prob += log_det_jacobian

    return outputs, log_prob



"""
Simple utils to save and load from disk.
"""
import joblib
import gzip
import pickle
import os
import tempfile
import tarfile
import zipfile
import logging
from urllib.request import urlretrieve
from typing import Any, Iterator, List, Optional, Tuple, Union, cast, IO

import pandas as pd
import numpy as np

import deepchem as dc

logger = logging.getLogger(__name__)


def pad_array(x: np.ndarray,
              shape: Union[Tuple, int],
              fill: float = 0.0,
              both: bool = False) -> np.ndarray:
  """
  Pad an array with a fill value.

  Parameters
  ----------
  x: np.ndarray
    A numpy array.
  shape: Tuple or int
    Desired shape. If int, all dimensions are padded to that size.
  fill: float, optional (default 0.0)
    The padded value.
  both: bool, optional (default False)
    If True, split the padding on both sides of each axis. If False,
    padding is applied to the end of each axis.

  Returns
  -------
  np.ndarray
    A padded numpy array
  """
  x = np.asarray(x)
  if not isinstance(shape, tuple):
    shape = tuple(shape for _ in range(x.ndim))
  pad = []
  for i in range(x.ndim):
    diff = shape[i] - x.shape[i]
    assert diff >= 0
    if both:
      a, b = divmod(diff, 2)
      b += a
      pad.append((a, b))
    else:
      pad.append((0, diff))
  pad = tuple(pad)  # type: ignore
  x = np.pad(x, pad, mode='constant', constant_values=fill)
  return x


def get_data_dir() -> str:
  """Get the DeepChem data directory.

  Returns
  -------
  str
    The default path to store DeepChem data. If you want to
    change this path, please set your own path to `DEEPCHEM_DATA_DIR`
    as an environment variable.
  """
  if 'DEEPCHEM_DATA_DIR' in os.environ:
    return os.environ['DEEPCHEM_DATA_DIR']
  return tempfile.gettempdir()


def download_url(url: str,
                 dest_dir: str = get_data_dir(),
                 name: Optional[str] = None):
  """Download a file to disk.

  Parameters
  ----------
  url: str
    The URL to download from
  dest_dir: str
    The directory to save the file in
  name: str
    The file name to save it as.  If omitted, it will try to extract a file name from the URL
  """
  if name is None:
    name = url
    if '?' in name:
      name = name[:name.find('?')]
    if '/' in name:
      name = name[name.rfind('/') + 1:]
  if not os.path.exists(dest_dir):
    os.makedirs(dest_dir)
  urlretrieve(url, os.path.join(dest_dir, name))


def untargz_file(file: str,
                 dest_dir: str = get_data_dir(),
                 name: Optional[str] = None):
  """Untar and unzip a .tar.gz file to disk.

  Parameters
  ----------
  file: str
    The filepath to decompress
  dest_dir: str
    The directory to save the file in
  name: str
    The file name to save it as.  If omitted, it will use the file name
  """
  if name is None:
    name = file
  tar = tarfile.open(name)
  tar.extractall(path=dest_dir)
  tar.close()


def unzip_file(file: str,
               dest_dir: str = get_data_dir(),
               name: Optional[str] = None):
  """Unzip a .zip file to disk.

  Parameters
  ----------
  file: str
    The filepath to decompress
  dest_dir: str
    The directory to save the file in
  name: str
    The directory name to unzip it to.  If omitted, it will use the file name
  """
  if name is None:
    name = file
  if dest_dir is None:
    dest_dir = os.path.join(get_data_dir, name)
  with zipfile.ZipFile(file, "r") as zip_ref:
    zip_ref.extractall(dest_dir)


class UniversalNamedTemporaryFile:
  """The implementation for cross platform NamedTemporaryFile.

  `tempfile.NamedTemporaryFile` causes a permission error on Windows.
  This implementation avoids the error, please see threads on the stackoverflow [1]_.

  References
  ----------
  .. [1] https://stackoverflow.com/questions/23212435/permission-denied-to-write-to-my-temporary-file
  """

  def __init__(self, mode='wb', delete=True):
    self._mode = mode
    self._delete = delete

  def __enter__(self):
    # Generate a random temporary file name
    file_name = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
    # Ensure the file is created
    open(file_name, "x").close()
    # Open the file in the given mode
    self._tempFile = open(file_name, self._mode)
    return self._tempFile

  def __exit__(self, exc_type, exc_val, exc_tb):
    self._tempFile.close()
    if self._delete:
      os.remove(self._tempFile.name)


def load_image_files(input_files: List[str]) -> np.ndarray:
  """Loads a set of images from disk.

  Parameters
  ----------
  input_files: List[str]
    List of image filenames.

  Returns
  -------
  np.ndarray
    A numpy array that contains loaded images. The shape is, `(N,...)`.

  Notes
  -----
  This method requires Pillow to be installed.
  The supported file types are PNG and TIF.
  """
  try:
    from PIL import Image
  except ModuleNotFoundError:
    raise ImportError("This function requires Pillow to be installed.")

  images = []
  for input_file in input_files:
    _, extension = os.path.splitext(input_file)
    extension = extension.lower()
    if extension == ".png":
      image = np.array(Image.open(input_file))
      images.append(image)
    elif extension == ".tif":
      im = Image.open(input_file)
      imarray = np.array(im)
      images.append(imarray)
    else:
      raise ValueError("Unsupported image filetype for %s" % input_file)
  return np.array(images)


def load_sdf_files(input_files: List[str],
                   clean_mols: bool = True,
                   tasks: List[str] = [],
                   shard_size: Optional[int] = None) -> Iterator[pd.DataFrame]:
  """Load SDF file into dataframe.

  Parameters
  ----------
  input_files: List[str]
    List of filenames
  clean_mols: bool, default True
    Whether to sanitize molecules.
  tasks: List[str], default []
    Each entry in `tasks` is treated as a property in the SDF file and is
    retrieved with `mol.GetProp(str(task))` where `mol` is the RDKit mol
    loaded from a given SDF entry.
  shard_size: int, default None
    The shard size to yield at one time.

  Returns
  -------
  Iterator[pd.DataFrame]
    Generator which yields the dataframe which is the same shard size.

  Notes
  -----
  This function requires RDKit to be installed.
  """
  try:
    from rdkit import Chem
  except ModuleNotFoundError:
    raise ImportError("This function requires RDKit to be installed.")

  df_rows = []
  for input_file in input_files:
    # Tasks are either in .sdf.csv file or in the .sdf file itself for QM9 dataset
    has_csv = os.path.isfile(input_file + ".csv")
    # Structures are stored in .sdf file
    logger.info("Reading structures from %s." % input_file)

    suppl = Chem.SDMolSupplier(str(input_file),
                               sanitize=clean_mols,
                               removeHs=False,
                               strictParsing=False)
    for ind, mol in enumerate(suppl):
      if mol is None:
        continue
      smiles = Chem.MolToSmiles(mol)
      df_row = [ind, smiles, mol]
      if not has_csv:  # Get task targets from .sdf file
        for task in tasks:
          df_row.append(mol.GetProp(str(task)))

      conf = mol.GetConformer()
      positions = conf.GetPositions()
      pos_x, pos_y, pos_z = zip(*positions)
      df_row.append(str(pos_x))
      df_row.append(str(pos_y))
      df_row.append(str(pos_z))
      df_rows.append(df_row)

      if shard_size is not None and len(df_rows) == shard_size:
        if has_csv:
          mol_df = pd.DataFrame(df_rows,
                                columns=('mol_id', 'smiles', 'mol', 'pos_x',
                                         'pos_y', 'pos_z'))
          raw_df = next(load_csv_files([input_file + ".csv"], shard_size=None))
          yield pd.concat([mol_df, raw_df], axis=1, join='inner')
        else:
          # Note: Here, the order of columns is based on the order in which the values
          # are appended to `df_row`. Since pos_x, pos_y, pos_z are appended after appending
          # tasks above, they occur after `tasks` here.
          # FIXME Ideally, we should use something like a dictionary here to keep it independent
          # of column ordering.
          mol_df = pd.DataFrame(df_rows,
                                columns=('mol_id', 'smiles', 'mol') +
                                tuple(tasks) + ('pos_x', 'pos_y', 'pos_z'))
          yield mol_df
        # Reset aggregator
        df_rows = []

    # Handle final leftovers for this file
    if len(df_rows) > 0:
      if has_csv:
        mol_df = pd.DataFrame(df_rows,
                              columns=('mol_id', 'smiles', 'mol', 'pos_x',
                                       'pos_y', 'pos_z'))
        raw_df = next(load_csv_files([input_file + ".csv"], shard_size=None))
        yield pd.concat([mol_df, raw_df], axis=1, join='inner')
      else:
        mol_df = pd.DataFrame(df_rows,
                              columns=('mol_id', 'smiles', 'mol') +
                              tuple(tasks) + ('pos_x', 'pos_y', 'pos_z'))
        yield mol_df
      df_rows = []


def load_csv_files(input_files: List[str],
                   shard_size: Optional[int] = None) -> Iterator[pd.DataFrame]:
  """Load data as pandas dataframe from CSV files.

  Parameters
  ----------
  input_files: List[str]
    List of filenames
  shard_size: int, default None
    The shard size to yield at one time.

  Returns
  -------
  Iterator[pd.DataFrame]
    Generator which yields the dataframe which is the same shard size.
  """
  # First line of user-specified CSV *must* be header.
  shard_num = 1
  for input_file in input_files:
    if shard_size is None:
      yield pd.read_csv(input_file)
    else:
      logger.info("About to start loading CSV from %s" % input_file)
      for df in pd.read_csv(input_file, chunksize=shard_size):
        logger.info("Loading shard %d of size %s." %
                    (shard_num, str(shard_size)))
        df = df.replace(np.nan, str(""), regex=True)
        shard_num += 1
        yield df


def load_json_files(input_files: List[str],
                    shard_size: Optional[int] = None) -> Iterator[pd.DataFrame]:
  """Load data as pandas dataframe.

  Parameters
  ----------
  input_files: List[str]
    List of json filenames.
  shard_size: int, default None
    Chunksize for reading json files.

  Returns
  -------
  Iterator[pd.DataFrame]
    Generator which yields the dataframe which is the same shard size.

  Notes
  -----
  To load shards from a json file into a Pandas dataframe, the file
  must be originally saved with ``df.to_json('filename.json', orient='records', lines=True)``
  """
  shard_num = 1
  for input_file in input_files:
    if shard_size is None:
      yield pd.read_json(input_file, orient='records', lines=True)
    else:
      logger.info("About to start loading json from %s." % input_file)
      for df in pd.read_json(input_file,
                             orient='records',
                             chunksize=shard_size,
                             lines=True):
        logger.info("Loading shard %d of size %s." %
                    (shard_num, str(shard_size)))
        df = df.replace(np.nan, str(""), regex=True)
        shard_num += 1
        yield df


def load_pickle_file(input_file: str) -> Any:
  """Load from single, possibly gzipped, pickle file.

  Parameters
  ----------
  input_file: str
    The filename of pickle file. This function can load from
    gzipped pickle file like `XXXX.pkl.gz`.

  Returns
  -------
  Any
    The object which is loaded from the pickle file.
  """
  if ".gz" in input_file:
    with gzip.open(input_file, "rb") as unzipped_file:
      return pickle.load(cast(IO[bytes], unzipped_file))
  else:
    with open(input_file, "rb") as opened_file:
      return pickle.load(opened_file)


def load_pickle_files(input_files: List[str]) -> Iterator[Any]:
  """Load dataset from pickle files.

  Parameters
  ----------
  input_files: List[str]
    The list of filenames of pickle file. This function can load from
    gzipped pickle file like `XXXX.pkl.gz`.

  Returns
  -------
  Iterator[Any]
    Generator which yields the objects which is loaded from each pickle file.
  """
  for input_file in input_files:
    yield load_pickle_file(input_file)


def load_data(input_files: List[str],
              shard_size: Optional[int] = None) -> Iterator[Any]:
  """Loads data from files.

  Parameters
  ----------
  input_files: List[str]
    List of filenames.
  shard_size: int, default None
    Size of shard to yield

  Returns
  -------
  Iterator[Any]
    Iterator which iterates over provided files.

  Notes
  -----
  The supported file types are SDF, CSV and Pickle.
  """
  if len(input_files) == 0:
    raise ValueError("The length of `filenames` must be more than 1.")

  file_type = _get_file_type(input_files[0])
  if file_type == "sdf":
    if shard_size is not None:
      logger.info("Ignoring shard_size for sdf input.")
    for value in load_sdf_files(input_files):
      yield value
  elif file_type == "csv":
    for value in load_csv_files(input_files, shard_size):
      yield value
  elif file_type == "pickle":
    if shard_size is not None:
      logger.info("Ignoring shard_size for pickle input.")
    for value in load_pickle_files(input_files):
      yield value


def _get_file_type(input_file: str) -> str:
  """Get type of input file. Must be csv/pkl/sdf/joblib file."""
  filename, file_extension = os.path.splitext(input_file)
  # If gzipped, need to compute extension again
  if file_extension == ".gz":
    filename, file_extension = os.path.splitext(filename)
  if file_extension == ".csv":
    return "csv"
  elif file_extension == ".pkl":
    return "pickle"
  elif file_extension == ".joblib":
    return "joblib"
  elif file_extension == ".sdf":
    return "sdf"
  else:
    raise ValueError("Unrecognized extension %s" % file_extension)


def save_to_disk(dataset: Any, filename: str, compress: int = 3):
  """Save a dataset to file.

  Parameters
  ----------
  dataset: str
    A data saved
  filename: str
    Path to save data.
  compress: int, default 3
    The compress option when dumping joblib file.
  """
  if filename.endswith('.joblib'):
    joblib.dump(dataset, filename, compress=compress)
  elif filename.endswith('.npy'):
    np.save(filename, dataset)
  else:
    raise ValueError("Filename with unsupported extension: %s" % filename)


def load_from_disk(filename: str) -> Any:
  """Load a dataset from file.

  Parameters
  ----------
  filename: str
    A filename you want to load data.

  Returns
  -------
  Any
    A loaded object from file.
  """
  name = filename
  if os.path.splitext(name)[1] == ".gz":
    name = os.path.splitext(name)[0]
  extension = os.path.splitext(name)[1]
  if extension == ".pkl":
    return load_pickle_file(filename)
  elif extension == ".joblib":
    return joblib.load(filename)
  elif extension == ".csv":
    # First line of user-specified CSV *must* be header.
    df = pd.read_csv(filename, header=0)
    df = df.replace(np.nan, str(""), regex=True)
    return df
  elif extension == ".npy":
    return np.load(filename, allow_pickle=True)
  else:
    raise ValueError("Unrecognized filetype for %s" % filename)


def load_dataset_from_disk(
    save_dir: str
) -> Tuple[bool, Optional[Tuple["dc.data.DiskDataset", "dc.data.DiskDataset",
                                "dc.data.DiskDataset"]],
           List["dc.trans.Transformer"]]:
  """Loads MoleculeNet train/valid/test/transformers from disk.

  Expects that data was saved using `save_dataset_to_disk` below. Expects the
  following directory structure for `save_dir`:
  save_dir/
    |
    ---> train_dir/
    |
    ---> valid_dir/
    |
    ---> test_dir/
    |
    ---> transformers.pkl

  Parameters
  ----------
  save_dir: str
    Directory name to load datasets.

  Returns
  -------
  loaded: bool
    Whether the load succeeded
  all_dataset: Tuple[DiskDataset, DiskDataset, DiskDataset]
    The train, valid, test datasets
  transformers: Transformer
    The transformers used for this dataset

  See Also
  --------
  save_dataset_to_disk
  """

  train_dir = os.path.join(save_dir, "train_dir")
  valid_dir = os.path.join(save_dir, "valid_dir")
  test_dir = os.path.join(save_dir, "test_dir")
  if not os.path.exists(train_dir) or not os.path.exists(
      valid_dir) or not os.path.exists(test_dir):
    return False, None, list()
  loaded = True
  train = dc.data.DiskDataset(train_dir)
  valid = dc.data.DiskDataset(valid_dir)
  test = dc.data.DiskDataset(test_dir)
  train.memory_cache_size = 40 * (1 << 20)  # 40 MB
  all_dataset = (train, valid, test)
  transformers = load_transformers(save_dir)
  return loaded, all_dataset, transformers


def save_dataset_to_disk(save_dir: str, train: "dc.data.DiskDataset",
                         valid: "dc.data.DiskDataset",
                         test: "dc.data.DiskDataset",
                         transformers: List["dc.trans.Transformer"]):
  """Utility used by MoleculeNet to save train/valid/test datasets.

  This utility function saves a train/valid/test split of a dataset along
  with transformers in the same directory. The saved datasets will take the
  following structure:
  save_dir/
    |
    ---> train_dir/
    |
    ---> valid_dir/
    |
    ---> test_dir/
    |
    ---> transformers.pkl

  Parameters
  ----------
  save_dir: str
    Directory name to save datasets to.
  train: DiskDataset
    Training dataset to save.
  valid: DiskDataset
    Validation dataset to save.
  test: DiskDataset
    Test dataset to save.
  transformers: List[Transformer]
    List of transformers to save to disk.

  See Also
  --------
  load_dataset_from_disk
  """
  train_dir = os.path.join(save_dir, "train_dir")
  valid_dir = os.path.join(save_dir, "valid_dir")
  test_dir = os.path.join(save_dir, "test_dir")
  train.move(train_dir)
  valid.move(valid_dir)
  test.move(test_dir)
  save_transformers(save_dir, transformers)


def load_transformers(save_dir: str) -> List["dc.trans.Transformer"]:
  """Load the transformers for a MoleculeNet dataset from disk."""
  with open(os.path.join(save_dir, "transformers.pkl"), 'rb') as f:
    return pickle.load(f)


def save_transformers(save_dir: str,
                      transformers: List["dc.trans.Transformer"]):
  """Save the transformers for a MoleculeNet dataset to disk."""
  with open(os.path.join(save_dir, "transformers.pkl"), 'wb') as f:
    pickle.dump(transformers, f)


def remove_missing_entries(dataset):
  """Remove missing entries.

  Some of the datasets have missing entries that sneak in as zero'd out
  feature vectors. Get rid of them.
  """
  for i, (X, y, w, ids) in enumerate(dataset.itershards()):
    available_rows = X.any(axis=1)
    logger.info("Shard %d has %d missing entries." %
                (i, np.count_nonzero(~available_rows)))
    X = X[available_rows]
    y = y[available_rows]
    w = w[available_rows]
    ids = ids[available_rows]
    dataset.set_shard(i, X, y, w, ids)
